//
//  main.c
//  AI003
//
//  Created by 西田　耀 on 13/01/28.
//  Copyright (c) 2013年 Hikaru Nishida. All rights reserved.
//

//
//Include headers
//

#include <stdio.h>
#include "core.h"

//
//Define static values
//

AI_WorkingSet WorkingSet;

int main(int argc, const char * argv[])
{
    //エントリポイント
    int i, i_max;
    CHNLIB_String *input, *temp;
    CHNLIB_UIPArray *separated;
    int passthink;
    
    CHNLIB_Environment_SetCurrentWorkingDirectory(argv[0]);
    
    CHNLIB_Debug("Hello, World.", __FILE__, __LINE__, __FUNCTION__);
    
    AI_System_InitializeSystemWorkingSet();
    
    AI_System_LoadMemory(AI_CONFIG_FILE_NAME);

    for(;;){
        passthink = False;
        input = CHNLIB_ReadLine(stdin);

        if(CHNLIB_String_CompareStringWithCString(input, "::")){
            //システムコマンド解釈
            passthink = True;
            separated = CHNLIB_UIPArray_Initialize();
            CHNLIB_UIPArray_GetSeparatedStringByUIPArray(&separated, WorkingSet.SystemWordList0, input);
            
            temp = CHNLIB_UIPArray_GetPointerByIndex(separated, 2);
            if(temp != NULL){
                if(CHNLIB_String_CompareStringWithCString(temp, "info")){
                    printf("AI internal information:\n");
                } else if(CHNLIB_String_CompareStringWithCString(temp, "exit")){
                    break;
                } else if(CHNLIB_String_CompareStringWithCString(temp, "wordlist")){
                    i_max = CHNLIB_UIPArray_GetNumberOfDatas(WorkingSet.RootWordList);
                    for(i = 0; i < i_max; i++){
                        printf("word%3d:%s\n", i, CHNLIB_String_GetReferencePointerOfCString(CHNLIB_UIPArray_GetPointerByIndex(WorkingSet.RootWordList, i)));
                    }
                } else{
                    passthink = False;
                }
            }
            
            CHNLIB_UIPArray_FreeSelectedAll(&separated);
        }
        
        if(!passthink){
            //AIへの入力
            separated = CHNLIB_UIPArray_Initialize();
            separated = AI_Think_SlideLookUpWordByHistory(input);
            i_max = CHNLIB_UIPArray_GetNumberOfDatas(separated);
            printf("Index(Decimal),CountOfContain(Decimal), String\n");
            for(i = 0; i < i_max; i++){
                printf("%d,%d,%s\n", i, CHNLIB_UIPArray_GetData32ByIndex(separated, i), CHNLIB_String_GetReferencePointerOfCString(CHNLIB_UIPArray_GetPointerByIndex(separated, i)));
            }
            CHNLIB_UIPArray_FreeAll(&separated);
            CHNLIB_UIPArray_AppendLast(&WorkingSet.InputHistory, CHNLIB_UIPArray_GetNumberOfDatas(WorkingSet.InputHistory), input);
        }
    }
    
    return 0;
}

CHNLIB_UIPArray *AI_Think_SlideLookUpWordByHistory(CHNLIB_String *input)
{
    //入力文字列と履歴文字列を照らし合わせ、単語の候補を抜き出す。
    CHNLIB_UIPArray *candidatewordlist;
    int i, i_max;
    int j, j_max;
    int k, k_max;
    int candidatelength, templength;
    const char *cstr_input, *cstr_history;
    
    candidatewordlist = CHNLIB_UIPArray_Initialize();
    
    i_max = CHNLIB_UIPArray_GetNumberOfDatas(WorkingSet.InputHistory);
    
    cstr_input = CHNLIB_String_GetReferencePointerOfCString(input);
    k_max = CHNLIB_CString_GetLength(cstr_input);

    for(k = 0; k < k_max; k++){
        //**UTF-8 only
        //UTF-8のマルチバイト部分でないことを確認
        if(CHNLIB_UTF8_GetCharacterType(cstr_input[k]) != 0){
            //****
            candidatelength = 0;
            for(i = 0; i < i_max; i++){
                cstr_history = CHNLIB_String_GetReferencePointerOfCString(CHNLIB_UIPArray_GetPointerByIndex(WorkingSet.InputHistory, i));
                j_max = CHNLIB_CString_GetLength(cstr_history);
                for(j = 0; j < j_max; j++){
                    templength = CHNLIB_CString_CompareString_LeftHand(&cstr_history[j], &cstr_input[k]);
                    if(templength > candidatelength && templength != (k_max - k)){
                        //前方一致の長さが、
                        //これまで見つかった単語よりも長く、かつ
                        //入力文字列の検索部分の全長ではない場合、
                        //単語の候補とする。
                        candidatelength = templength;
                    }
                }
            }
            if(candidatelength > 0){
                //AI_Memory_AddRootWordData(CHNLIB_String_ExtractByLength(input, k, candidatelength));
                CHNLIB_UIPArray_AppendLast_ProtectFromDuplication(&candidatewordlist, 0, CHNLIB_String_ExtractByLength(input, k, candidatelength), &AI_Memory_AddRootWordData_IsDuplicated);
            }
        }
    }
    
    //各候補単語が入力文字列にいくつ含まれているかをdata32に保存する。
    i_max = CHNLIB_UIPArray_GetNumberOfDatas(candidatewordlist);
    for(i = 0; i < i_max; i++){
        CHNLIB_UIPArray_SetData32ByIndex(candidatewordlist, i, CHNLIB_String_GetCountOfContain(input, CHNLIB_UIPArray_GetPointerByIndex(candidatewordlist, i)));
    }
    
    return candidatewordlist;
}

void AI_Memory_AddRootWordData(CHNLIB_String *tag)
{
    //WorkingSet.RootWordListに文字列を追加する。
    //重複がある場合は警告を出し、追加しない。
    int i, i_max;
    
    i_max = CHNLIB_UIPArray_GetNumberOfDatas(WorkingSet.RootWordList);
    for(i = 0; i < i_max; i++){
        if(CHNLIB_String_CompareString_Strict(CHNLIB_UIPArray_GetPointerByIndex(WorkingSet.RootWordList, i), tag)){
            CHNLIB_ReportError("Word[%s] has already existed.", __FILE__, __LINE__, __FUNCTION__, CHNLIB_String_GetReferencePointerOfCString(tag));
            return;
        }
    }
    
    CHNLIB_UIPArray_AppendLast(&WorkingSet.RootWordList, i_max, tag);
    
    CHNLIB_UIPArray_AppendLast_ProtectFromDuplication(&WorkingSet.RootWordList, CHNLIB_UIPArray_GetNumberOfDatas(WorkingSet.RootWordList), tag, &AI_Memory_AddRootWordData_IsDuplicated);
    
    return;
}

int AI_Memory_AddRootWordData_IsDuplicated(const void *listtag, const void *newtag)
{
    //CHNLIB_UIPArray_AppendLast_ProtectFromDuplicationに渡す関数ポインタ。
    if(CHNLIB_StructureHeader_GetTypeID(listtag) != CHNLIB_STRUCT_ID_String){
        return False;
    }
    if(CHNLIB_String_CompareString_Strict(listtag, newtag)){
        CHNLIB_ReportError("Word[%s] has already existed.", __FILE__, __LINE__, __FUNCTION__, CHNLIB_String_GetReferencePointerOfCString(newtag));
        return True;
    }
    return False;
}

void AI_System_InitializeSystemWorkingSet(void)
{
    //WorkingSetを初期化する。
    WorkingSet.SystemWordList0 = CHNLIB_UIPArray_Initialize();
    CHNLIB_UIPArray_AppendLast(&WorkingSet.SystemWordList0, True, CHNLIB_String_Initialize(":"));
    CHNLIB_UIPArray_AppendLast(&WorkingSet.SystemWordList0, True, CHNLIB_String_Initialize(","));
    
    WorkingSet.InputHistory = CHNLIB_UIPArray_Initialize();
    
    return;
}

void AI_System_LoadMemory(const char configfilename[])
{
    //指定されたファイル名のテキストファイルから設定を読み出す。
    FILE *fp;
    CHNLIB_String *line, *currentfilename;
    int i;
    
    fp = fopen(configfilename, "rb");
    currentfilename = CHNLIB_ReadLine(fp);
    fclose(fp);
    
    fp = fopen(CHNLIB_String_GetReferencePointerOfCString(currentfilename), "rb");
    
    WorkingSet.RootWordList = CHNLIB_UIPArray_Initialize();
    for(i = 1; ;i++){
        line = CHNLIB_ReadLine(fp);
        if(line == NULL){
            break;
        }
        CHNLIB_UIPArray_AppendLast(&WorkingSet.RootWordList, i, line);
        //CHNLIB_String_Free(line);リストに登録している実体なので解放の必要なし！
    }
    fclose(fp);
    CHNLIB_String_Free(currentfilename);
    
    //CHNLIB_Debug_PrintStructureData(WorkingSet.RootWordList, 0);
    
    return;
}
