//
//  chnlib.c
//  AI003
//
//  Created by 西田　耀 on 13/01/28.
//  Copyright (c) 2013年 Hikaru Nishida. All rights reserved.
//

//
//Include headers
//

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "chnlib.h"

//
//Define static values
//
int CHNLIB_Debug_PrintStructureData_RecursiveCounter;

//
//Declare internal functions
//
const char *CHNLIB_Debug_Internal_GetStructTypeNameByID(uint typeid);
void CHNLIB_Debug_Internal_PrintStructureDataSub(void *structure, uint level);
void CHNLIB_Debug_Internal_PrintIndent(int level);

//
//Define types
//

//
//Functions
//

void CHNLIB_Debug(const char format[], const char filename[], int line, const char funcname[], ...)
{
    char s[CHNLIB_MAX_STRING_LENGTH];
    va_list varg;
    
    fprintf(stdout, "CHNLIB:Debug:%s@%s:%d:", funcname, filename, line);
    
    va_start(varg, funcname);
    vsnprintf(s, sizeof(s), format, varg);
    va_end(varg);
    
    fputs(s, stdout);
    fputs("\n", stdout);
	
	fflush(stdout);
	
    return;
}

void CHNLIB_Debug_PrintStructureData(void *structure, uint level)
{
    //level == 0:全階層
    //level >= 1:level階層まで

    CHNLIB_Debug_PrintStructureData_RecursiveCounter = level - 1;
    
    CHNLIB_Debug_Internal_PrintStructureDataSub(structure, 0);
    
    return;
}

void CHNLIB_ReportError(const char format[], const char filename[], int line, const char funcname[], ...)
{
    char s[CHNLIB_MAX_STRING_LENGTH];
    va_list varg;
    
    fprintf(stderr, "CHNLIB:Error:%s@%s:%d:", funcname, filename, line);
    
    va_start(varg, funcname);
    vsnprintf(s, sizeof(s), format, varg);
    va_end(varg);
    
    fputs(s, stderr);
	fputs("\n", stderr);
    
    return;
}

void CHNLIB_StructureHeader_Initialize(CHNLIB_StructureHeader *header, uint typeid)
{
    if(header == NULL){
        CHNLIB_ReportError("Null structure.", __FILE__, __LINE__, __FUNCTION__);
        return;
    }
    
    header->signature = CHNLIB_STRUCTURE_SIGNATURE;
    header->typeid = typeid;
    return;
}

uint CHNLIB_StructureHeader_GetTypeID(const void *structure)
{
    CHNLIB_StructureHeader *strhead;
    
    if(structure == NULL){
        return CHNLIB_STRUCT_ID_Null;
    }
    
    strhead = (CHNLIB_StructureHeader *)structure;
    if(strhead->signature == CHNLIB_STRUCTURE_SIGNATURE){
        return strhead->typeid;
    }
    return CHNLIB_STRUCT_ID_Null;
}

void *CHNLIB_System_AllocateMemory_Strict(int size, const char filename[], int line, const char funcname[])
{
    void *p;
    
    p = malloc(size);
    if(p == NULL){
        CHNLIB_ReportError("An attempt to allocate memory %d bytes in %s@%s:%d: is failed. Abort.", __FILE__, __LINE__, __FUNCTION__, size, filename, line, funcname);
        exit(EXIT_FAILURE);
    }
    
#ifdef DEBUG_MEMORY_ALLOCATION
    CHNLIB_Debug("Request allocation %d bytes ->[%p].", __FILE__, __LINE__, __FUNCTION__, size, p);
#endif
    
    memset(p, 0, size);

    return p;
}

void CHNLIB_System_FreeMemory(void *p, const char filename[], int line, const char funcname[])
{
    if(p == NULL){
        CHNLIB_ReportError("An attempt to free NULL memory in %s@%s:%d:. Ignore.", __FILE__, __LINE__, __FUNCTION__, filename, line, funcname);
        return;
    }
    
#ifdef DEBUG_MEMORY_ALLOCATION
    CHNLIB_Debug("Request ƒree memory [%p].", __FILE__, __LINE__, __FUNCTION__, p);
#endif
    
    free(p);
    
    return;

}

//
//Internal functions
//

const char *CHNLIB_Debug_Internal_GetStructTypeNameByID(uint typeid)
{
    switch(typeid){
        case CHNLIB_STRUCT_ID_Null:
            return "Null";
        case CHNLIB_STRUCT_ID_UIPArray:
            return "UIPArray";
        case CHNLIB_STRUCT_ID_String:
            return "String";
    }
    return "Unknown";
}

void CHNLIB_Debug_Internal_PrintStructureDataSub(void *structure, uint level)
{
    int typeid, i, j;
    void *p;
    
    typeid = CHNLIB_StructureHeader_GetTypeID(structure);
    
    switch(typeid){
        case CHNLIB_STRUCT_ID_UIPArray:
            i = CHNLIB_UIPArray_GetNumberOfDatas(structure);
            CHNLIB_Debug_Internal_PrintIndent(level);
            CHNLIB_Debug("%s[%p]:Number of datas=%d", "", 0, "", CHNLIB_Debug_Internal_GetStructTypeNameByID(typeid), structure, i);
            for(j = 0; j < i; j++){
                p = CHNLIB_UIPArray_GetPointerByIndex(structure, j);
                CHNLIB_Debug_Internal_PrintIndent(level);
                CHNLIB_Debug("%s[%p]:[%d]=(%u,[%p])", "", 0, "", CHNLIB_Debug_Internal_GetStructTypeNameByID(typeid), structure, j, CHNLIB_UIPArray_GetData32ByIndex(structure, j), p);
                if(CHNLIB_Debug_PrintStructureData_RecursiveCounter != 0){
                    CHNLIB_Debug_Internal_PrintStructureDataSub(p, level + 1);
                }
            }
            break;
        case CHNLIB_STRUCT_ID_String:
            CHNLIB_Debug_Internal_PrintIndent(level);
            CHNLIB_Debug("%s[%p]:[%s]", "", 0, "", CHNLIB_Debug_Internal_GetStructTypeNameByID(typeid), structure, CHNLIB_String_GetReferencePointerOfCString(structure));
            break;
        default:
            CHNLIB_Debug_Internal_PrintIndent(level);
            CHNLIB_Debug("%s[%p]:NULL structure or not implemented.", "", 0, "", CHNLIB_Debug_Internal_GetStructTypeNameByID(typeid), structure);
            break;
    }
    return;
}

void CHNLIB_Debug_Internal_PrintIndent(int level)
{
    int i;
    
    for(i = 0; i < level; i++){
        fputs("    ", stdout);
    }
	
    return;
}
